@@ -4,8 +4,7 @@ import copy
 
 import awkward as ak
 from awkward._v2.index import Index
-from awkward._v2._slicing import NestedIndexError
-from awkward._v2.contents.content import Content
+from awkward._v2.contents.content import Content, unset
 from awkward._v2.forms.indexedform import IndexedForm
 from awkward._v2.forms.form import _parameters_equal
 
@@ -16,6 +15,22 @@ numpy = ak.nplike.Numpy.instance()
 class IndexedArray(Content):
     is_IndexedType = True
 
+    def copy(
+        self,
+        index=unset,
+        content=unset,
+        identifier=unset,
+        parameters=unset,
+        nplike=unset,
+    ):
+        return IndexedArray(
+            self._index if index is unset else index,
+            self._content if content is unset else content,
+            self._identifier if identifier is unset else identifier,
+            self._parameters if parameters is unset else parameters,
+            self._nplike if nplike is unset else nplike,
+        )
+
     def __init__(self, index, content, identifier=None, parameters=None, nplike=None):
         if not (
             isinstance(index, Index)
@@ -26,14 +41,18 @@ class IndexedArray(Content):
                 np.dtype(np.int64),
             )
         ):
-            raise TypeError(
-                "{} 'index' must be an Index with dtype in (int32, uint32, int64), "
-                "not {}".format(type(self).__name__, repr(index))
+            raise ak._v2._util.error(
+                TypeError(
+                    "{} 'index' must be an Index with dtype in (int32, uint32, int64), "
+                    "not {}".format(type(self).__name__, repr(index))
+                )
             )
         if not isinstance(content, Content):
-            raise TypeError(
-                "{} 'content' must be a Content subtype, not {}".format(
-                    type(self).__name__, repr(content)
+            raise ak._v2._util.error(
+                TypeError(
+                    "{} 'content' must be a Content subtype, not {}".format(
+                        type(self).__name__, repr(content)
+                    )
                 )
             )
         if nplike is None:
@@ -140,7 +159,7 @@ class IndexedArray(Content):
         if where < 0:
             where += self.length
         if self._nplike.known_shape and not 0 <= where < self.length:
-            raise NestedIndexError(self, where)
+            raise ak._v2._util.indexerror(self, where)
         return self._content._getitem_at(self._index[where])
 
     def _getitem_range(self, where):
@@ -175,28 +194,25 @@ class IndexedArray(Content):
             self._nplike,
         )
 
-    def _carry(self, carry, allow_lazy, exception):
+    def _carry(self, carry, allow_lazy):
         assert isinstance(carry, ak._v2.index.Index)
 
         try:
             nextindex = self._index[carry.data]
         except IndexError as err:
-            if issubclass(exception, NestedIndexError):
-                raise exception(self, carry.data, str(err))
-            else:
-                raise exception(str(err))
+            raise ak._v2._util.indexerror(self, carry.data, str(err)) from err
 
         return IndexedArray(
             nextindex,
             self._content,
-            self._carry_identifier(carry, exception),
+            self._carry_identifier(carry),
             self._parameters,
             self._nplike,
         )
 
     def _getitem_next_jagged_generic(self, slicestarts, slicestops, slicecontent, tail):
         if self._nplike.known_shape and slicestarts.length != self.length:
-            raise NestedIndexError(
+            raise ak._v2._util.indexerror(
                 self,
                 ak._v2.contents.ListArray(
                     slicestarts, slicestops, slicecontent, None, None, self._nplike
@@ -218,10 +234,11 @@ class IndexedArray(Content):
                 self._index.data,
                 self._index.length,
                 self._content.length,
-            )
+            ),
+            slicer=ak._v2.contents.ListArray(slicestarts, slicestops, slicecontent),
         )
         # an eager carry (allow_lazy = false) to avoid infinite loop (unproven)
-        next = self._content._carry(nextcarry, False, NestedIndexError)
+        next = self._content._carry(nextcarry, False)
         return next._getitem_next_jagged(slicestarts, slicestops, slicecontent, tail)
 
     def _getitem_next_jagged(self, slicestarts, slicestops, slicecontent, tail):
@@ -233,7 +250,9 @@ class IndexedArray(Content):
         if head == ():
             return self
 
-        elif isinstance(head, (int, slice, ak._v2.index.Index64)):
+        elif isinstance(
+            head, (int, slice, ak._v2.index.Index64, ak._v2.contents.ListOffsetArray)
+        ):
             nexthead, nexttail = ak._v2._slicing.headtail(tail)
 
             nextcarry = ak._v2.index.Index64.empty(self._index.length, self._nplike)
@@ -250,10 +269,11 @@ class IndexedArray(Content):
                     self._index.data,
                     self._index.length,
                     self._content.length,
-                )
+                ),
+                slicer=head,
             )
 
-            next = self._content._carry(nextcarry, False, NestedIndexError)
+            next = self._content._carry(nextcarry, False)
             return next._getitem_next(head, tail, advanced)
 
         elif ak._util.isstr(head):
@@ -268,21 +288,20 @@ class IndexedArray(Content):
         elif head is Ellipsis:
             return self._getitem_next_ellipsis(tail, advanced)
 
-        elif isinstance(head, ak._v2.contents.ListOffsetArray):
-            raise NotImplementedError
-
         elif isinstance(head, ak._v2.contents.IndexedOptionArray):
             return self._getitem_next_missing(head, tail, advanced)
 
         else:
-            raise AssertionError(repr(head))
+            raise ak._v2._util.error(AssertionError(repr(head)))
 
     def project(self, mask=None):
         if mask is not None:
             if self._nplike.known_shape and self._index.length != mask.length:
-                raise ValueError(
-                    "mask length ({}) is not equal to {} length ({})".format(
-                        mask.length(), type(self).__name__, self._index.length
+                raise ak._v2._util.error(
+                    ValueError(
+                        "mask length ({}) is not equal to {} length ({})".format(
+                            mask.length(), type(self).__name__, self._index.length
+                        )
                     )
                 )
             nextindex = ak._v2.index.Index64.empty(self._index.length, self._nplike)
@@ -330,7 +349,7 @@ class IndexedArray(Content):
                     self._content.length,
                 )
             )
-            return self._content._carry(nextcarry, False, NestedIndexError)
+            return self._content._carry(nextcarry, False)
 
     def simplify_optiontype(self):
         if isinstance(
@@ -416,18 +435,18 @@ class IndexedArray(Content):
     def num(self, axis, depth=0):
         posaxis = self.axis_wrap_if_negative(axis)
         if posaxis == depth:
-            out = ak._v2.index.Index64.empty(1, self._nplike)
-            out[0] = self.length
-            return ak._v2.contents.numpyarray.NumpyArray(out, None, None, self._nplike)[
-                0
-            ]
+            out = self.length
+            if ak._v2._util.isint(out):
+                return np.int64(out)
+            else:
+                return out
         else:
             return self.project().num(posaxis, depth)
 
     def _offsets_and_flattened(self, axis, depth):
         posaxis = self.axis_wrap_if_negative(axis)
         if posaxis == depth:
-            raise np.AxisError(self, "axis=0 not allowed for flatten")
+            raise ak._v2._util.error(np.AxisError("axis=0 not allowed for flatten"))
 
         else:
             return self.project()._offsets_and_flattened(posaxis, depth)
@@ -462,8 +481,10 @@ class IndexedArray(Content):
 
     def _merging_strategy(self, others):
         if len(others) == 0:
-            raise ValueError(
-                "to merge this array with 'others', at least one other must be provided"
+            raise ak._v2._util.error(
+                ValueError(
+                    "to merge this array with 'others', at least one other must be provided"
+                )
             )
 
         head = [self]
@@ -558,9 +579,10 @@ class IndexedArray(Content):
         nextindex = ak._v2.index.Index64.empty(total_length, self._nplike)
         parameters = self._parameters
 
+        parameters = self._parameters
         for array in head:
             parameters = ak._v2._util.merge_parameters(
-                self._parameters, array._parameters, True
+                parameters, array._parameters, True
             )
 
             if isinstance(
@@ -630,31 +652,34 @@ class IndexedArray(Content):
         else:
             return reversed.mergemany(tail[1:])
 
-        raise NotImplementedError(
-            "not implemented: " + type(self).__name__ + " ::mergemany"
+        raise ak._v2._util.error(
+            NotImplementedError(
+                "not implemented: " + type(self).__name__ + " ::mergemany"
+            )
         )
 
-    def fillna(self, value):
+    def fill_none(self, value):
         if value.nplike.known_shape and value.length != 1:
-            raise ValueError(f"fillna value length ({value.length}) is not equal to 1")
-
+            raise ak._v2._util.error(
+                ValueError(f"fill_none value length ({value.length}) is not equal to 1")
+            )
         return IndexedArray(
             self._index,
-            self._content.fillna(value),
+            self._content.fill_none(value),
             None,
             self._parameters,
             self._nplike,
         )
 
-    def _localindex(self, axis, depth):
+    def _local_index(self, axis, depth):
         posaxis = self.axis_wrap_if_negative(axis)
         if posaxis == depth:
-            return self._localindex_axis0()
+            return self._local_index_axis0()
         else:
-            return self.project()._localindex(posaxis, depth)
+            return self.project()._local_index(posaxis, depth)
 
     def _unique_index(self, index, sorted=True):
-        next = ak._v2.index.Index64.empty(self.length, self._nplike)
+        next = ak._v2.index.Index64.zeros(self.length, self._nplike)
         length = ak._v2.index.Index64.zeros(1, self._nplike)
 
         if not sorted:
@@ -727,7 +752,10 @@ class IndexedArray(Content):
 
         nextindex = self._unique_index(self._index)
 
-        next = self._content._carry(nextindex, False, NestedIndexError)
+        if len(nextindex) != len(self._index):
+            return False
+
+        next = self._content._carry(nextindex, False)
         return next._is_unique(negaxis, starts, parents, outlength)
 
     def _unique(self, negaxis, starts, parents, outlength):
@@ -767,7 +795,7 @@ class IndexedArray(Content):
                 index_length,
             )
         )
-        next = self._content._carry(nextcarry, False, NestedIndexError)
+        next = self._content._carry(nextcarry, False)
         unique = next._unique(
             negaxis,
             starts,
@@ -814,12 +842,14 @@ class IndexedArray(Content):
             if isinstance(unique, ak._v2.contents.RegularArray):
                 unique = unique.toListOffsetArray64(True)
 
-            elif isinstance(unique, ak._v2.contents.ListOffsetArray):
+            if isinstance(unique, ak._v2.contents.ListOffsetArray):
                 if starts.nplike.known_data and starts.length > 0 and starts[0] != 0:
-                    raise AssertionError(
-                        "reduce_next with unbranching depth > negaxis expects a "
-                        "ListOffsetArray64 whose offsets start at zero ({})".format(
-                            starts[0]
+                    raise ak._v2._util.error(
+                        AssertionError(
+                            "reduce_next with unbranching depth > negaxis expects a "
+                            "ListOffsetArray64 whose offsets start at zero ({})".format(
+                                starts[0]
+                            )
                         )
                     )
 
@@ -858,7 +888,8 @@ class IndexedArray(Content):
 
             elif isinstance(unique, ak._v2.contents.NumpyArray):
                 nextoutindex = ak._v2.index.Index64(
-                    self._nplike.arange(unique.length, dtype=np.int64)
+                    self._nplike.index_nplike.arange(unique.length, dtype=np.int64),
+                    nplike=self.nplike,
                 )
                 out = ak._v2.contents.IndexedOptionArray(
                     nextoutindex,
@@ -870,7 +901,7 @@ class IndexedArray(Content):
 
                 return out
 
-        raise NotImplementedError
+        raise ak._v2._util.error(NotImplementedError)
 
     def _argsort_next(
         self,
@@ -884,12 +915,7 @@ class IndexedArray(Content):
         kind,
         order,
     ):
-        if self._index.length == 0:
-            return ak._v2.contents.NumpyArray(
-                self._nplike.empty(0, np.int64), None, None, self._nplike
-            )
-
-        next = self._content._carry(self._index, False, NestedIndexError)
+        next = self._content._carry(self._index, False)
         return next._argsort_next(
             negaxis,
             starts,
@@ -913,7 +939,7 @@ class IndexedArray(Content):
         kind,
         order,
     ):
-        next = self._content._carry(self._index, False, NestedIndexError)
+        next = self._content._carry(self._index, False)
         return next._sort_next(
             negaxis,
             starts,
@@ -944,6 +970,7 @@ class IndexedArray(Content):
         outlength,
         mask,
         keepdims,
+        behavior,
     ):
         branch, depth = self.branch_depth
 
@@ -976,7 +1003,7 @@ class IndexedArray(Content):
                 index_length,
             )
         )
-        next = self._content._carry(nextcarry, False, NestedIndexError)
+        next = self._content._carry(nextcarry, False)
         nextshifts = None
         out = next._reduce_next(
             reducer,
@@ -987,20 +1014,29 @@ class IndexedArray(Content):
             outlength,
             mask,
             keepdims,
+            behavior,
         )
 
+        # If we are reducing the contents of this layout,
+        # then we do NOT want to return an optional layout
         if not branch and negaxis == depth:
             return out
         else:
             if isinstance(out, ak._v2.contents.RegularArray):
                 out = out.toListOffsetArray64(True)
 
-            elif isinstance(out, ak._v2.contents.ListOffsetArray):
+            # If the result of `_reduce_next` is a list, and we're not applying at this
+            # depth, then it will have offsets given by the boundaries in parents.
+            # This means that we need to look at the _contents_ to which the `outindex`
+            # belongs to add the new index
+            if isinstance(out, ak._v2.contents.ListOffsetArray):
                 if starts.nplike.known_data and starts.length > 0 and starts[0] != 0:
-                    raise AssertionError(
-                        "reduce_next with unbranching depth > negaxis expects a "
-                        "ListOffsetArray64 whose offsets start at zero ({})".format(
-                            starts[0]
+                    raise ak._v2._util.error(
+                        AssertionError(
+                            "reduce_next with unbranching depth > negaxis expects a "
+                            "ListOffsetArray64 whose offsets start at zero ({})".format(
+                                starts[0]
+                            )
                         )
                     )
 
@@ -1021,7 +1057,7 @@ class IndexedArray(Content):
                     )
                 )
 
-                tmp = ak._v2.contents.IndexedArray(
+                inner = ak._v2.contents.IndexedArray(
                     outindex,
                     out._content,
                     None,
@@ -1031,20 +1067,22 @@ class IndexedArray(Content):
 
                 return ak._v2.contents.ListOffsetArray(
                     outoffsets,
-                    tmp,
+                    inner,
                     None,
                     None,
                     self._nplike,
                 )
 
             else:
-                raise AssertionError(
-                    "reduce_next with unbranching depth > negaxis is only "
-                    "expected to return RegularArray or ListOffsetArray64; "
-                    "instead, it returned " + out.classname
+                raise ak._v2._util.error(
+                    AssertionError(
+                        "reduce_next with unbranching depth > negaxis is only "
+                        "expected to return RegularArray or ListOffsetArray64; "
+                        "instead, it returned " + out.classname
+                    )
                 )
 
-    def _validityerror(self, path):
+    def _validity_error(self, path):
         error = self._nplike["awkward_IndexedArray_validity", self.index.dtype.type](
             self.index.data, self.index.length, self._content.length, False
         )
@@ -1072,7 +1110,7 @@ class IndexedArray(Content):
         ):
             return "{0} contains \"{1}\", the operation that made it might have forgotten to call 'simplify_optiontype()'"
         else:
-            return self._content.validityerror(path + ".content")
+            return self._content.validity_error(path + ".content")
 
     def _nbytes_part(self):
         result = self.index._nbytes_part() + self.content._nbytes_part()
@@ -1080,16 +1118,16 @@ class IndexedArray(Content):
             result = result + self.identifier._nbytes_part()
         return result
 
-    def _rpad(self, target, axis, depth, clip):
+    def _pad_none(self, target, axis, depth, clip):
         posaxis = self.axis_wrap_if_negative(axis)
         if posaxis == depth:
-            return self.rpad_axis0(target, clip)
+            return self.pad_none_axis0(target, clip)
         elif posaxis == depth + 1:
-            return self.project()._rpad(target, posaxis, depth, clip)
+            return self.project()._pad_none(target, posaxis, depth, clip)
         else:
             return ak._v2.contents.indexedarray.IndexedArray(
                 self._index,
-                self._content._rpad(target, posaxis, depth, clip),
+                self._content._pad_none(target, posaxis, depth, clip),
                 None,
                 self._parameters,
                 self._nplike,
@@ -1125,7 +1163,11 @@ class IndexedArray(Content):
             if options["extensionarray"]:
                 return ak._v2._connect.pyarrow.AwkwardArrowArray.from_storage(
                     ak._v2._connect.pyarrow.to_awkwardarrow_type(
-                        out.type, options["extensionarray"], mask_node, self
+                        out.type,
+                        options["extensionarray"],
+                        options["record_is_scalar"],
+                        mask_node,
+                        self,
                     ),
                     out,
                 )
@@ -1139,9 +1181,7 @@ class IndexedArray(Content):
                 # In that case, don't call self._content[index]; it's empty anyway.
                 next = self._content
             else:
-                next = self._content._carry(
-                    ak._v2.index.Index(index), False, IndexError
-                )
+                next = self._content._carry(ak._v2.index.Index(index), False)
 
             return next.merge_parameters(self._parameters)._to_arrow(
                 pyarrow, mask_node, validbytes, length, options
@@ -1154,15 +1194,28 @@ class IndexedArray(Content):
         return self.project()._completely_flatten(nplike, options)
 
     def _recursively_apply(
-        self, action, depth, depth_context, lateral_context, options
+        self, action, behavior, depth, depth_context, lateral_context, options
     ):
+        if (
+            self._nplike.known_shape
+            and self._nplike.known_data
+            and self._index.length != 0
+        ):
+            npindex = self._index.data
+            indexmin = npindex.min()
+            index = ak._v2.index.Index(npindex - indexmin, nplike=self._nplike)
+            content = self._content[indexmin : npindex.max() + 1]
+        else:
+            index, content = self._index, self._content
+
         if options["return_array"]:
 
             def continuation():
                 return IndexedArray(
-                    self._index,
-                    self._content._recursively_apply(
+                    index,
+                    content._recursively_apply(
                         action,
+                        behavior,
                         depth,
                         copy.copy(depth_context),
                         lateral_context,
@@ -1176,8 +1229,9 @@ class IndexedArray(Content):
         else:
 
             def continuation():
-                self._content._recursively_apply(
+                content._recursively_apply(
                     action,
+                    behavior,
                     depth,
                     copy.copy(depth_context),
                     lateral_context,
@@ -1190,6 +1244,8 @@ class IndexedArray(Content):
             depth_context=depth_context,
             lateral_context=lateral_context,
             continuation=continuation,
+            behavior=behavior,
+            nplike=self._nplike,
             options=options,
         )
 
@@ -1198,22 +1254,19 @@ class IndexedArray(Content):
         elif result is None:
             return continuation()
         else:
-            raise AssertionError(result)
+            raise ak._v2._util.error(AssertionError(result))
 
     def packed(self):
         return self.project().packed()
 
-    def _to_list(self, behavior):
-        out = self._to_list_custom(behavior)
+    def _to_list(self, behavior, json_conversions):
+        out = self._to_list_custom(behavior, json_conversions)
         if out is not None:
             return out
 
         index = self._index.raw(numpy)
-        content = self._content._to_list(behavior)
-        out = [None] * index.length
-        for i, ind in enumerate(index):
-            out[i] = content[ind]
-        return out
+        nextcontent = self._content._carry(ak._v2.index.Index(index), False)
+        return nextcontent._to_list(behavior, json_conversions)
 
     def _to_nplike(self, nplike):
         index = self._index._to_nplike(nplike)
@@ -1226,27 +1279,7 @@ class IndexedArray(Content):
             nplike=nplike,
         )
 
-    def _to_json(
-        self,
-        nan_string,
-        infinity_string,
-        minus_infinity_string,
-        complex_real_string,
-        complex_imag_string,
-    ):
-        out = self._to_json_custom()
-        if out is not None:
-            return out
-
-        index = self._index.raw(numpy)
-        content = self._content._to_json(
-            nan_string,
-            infinity_string,
-            minus_infinity_string,
-            complex_real_string,
-            complex_imag_string,
-        )
-        out = [None] * index.length
-        for i, ind in enumerate(index):
-            out[i] = content[ind]
-        return out
+    def _layout_equal(self, other, index_dtype=True, numpyarray=True):
+        return self.index.layout_equal(
+            other.index, index_dtype, numpyarray
+        ) and self.content.layout_equal(other.content, index_dtype, numpyarray)
